RaidGnome = LibStub("AceAddon-3.0"):NewAddon("RaidGnome", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "LibMMButton-1.0")

local TalentQuery = LibStub:GetLibrary("LibTalentQuery-1.0")
local L = LibStub("AceLocale-3.0"):GetLocale("RaidGnome", false)

local defaults = {
        char = {
        	sets = {},
        	verbose = true,
        	currentset = nil,
        	learnassistants = true,
        	learnset = true,
        	lootboss = "master",
        	lootdefault = "needbeforegreed",
        	assistants = {},
        }
}

local emptyset = {
	groups = {}
}

local raidcache = {}

local db

local WEIGHT_NAME	= 5
local WEIGHT_CLASS	= 1
local WEIGHT_SPEC	= 3

local MINIMUM_SCORE = 2

local loottypes = {
	freeforall = "Free for all", 
	roundrobin = "Rounde robin", 
	master = "Master loot", 
	group = "Group loot", 
	needbeforegreed = "Need before greed"
}


function RaidGnome:GetSetNames()
	local setnames = {}
	for i, set in ipairs(db.sets) do
		table.insert(setnames, set.name)
	end
	return setnames
end

-- options
RaidGnome.options = {
	type="group",
	name="RaidGnome",
	childGroups="tab",
	args={
		d = {
			type="description",
			name=L["RAIDGNOME_DESC"],
		},
		
		loot = {
			type="group",
			name=L["LOOTOPTIONS_NAME"],
			args = {
				
				d = {
					type="description",
					name=L["LOOTOPTIONS_DESC"],
					order=0,
				},
				
				lootdefault = {
					type="select",
					name=L["LOOTDEFAULT_NAME"],
					desc=L["LOOTDEFAULT_DESC"],
					order=10,
					values=loottypes,
					get=function() return db.lootdefault end,
					set=function(info, key) db.lootdefault = key end,
				},
				
				lootboss = {
					type="select",
					name=L["LOOTBOSS_NAME"],
					desc=L["LOOTBOSS_DESC"],
					order=11,
					values=loottypes,
					get=function() return db.lootboss end,
					set=function(info, key) db.lootboss = key end,
				},
				
			}
		},

		options = {
			type="group",
			name=L["GENERALOPTIONS_NAME"],
			args = {
				
				d = {
					type="description",
					name=L["GENERALOPTIONS_DESC"],
					order=0,
				},
				
				verbose = {
					type="toggle",
					name=L["VERBOSE_NAME"],
					desc=L["VERBOSE_DESC"],
					order=10,
					get=function() return db.verbose end,
					set=function() db.verbose = not db.verbose end,
				},
				
			}
		},
		
		assistants = {
			type="group",
			name=L["ASSISTANTS_NAME"],
			args = {
				
				d = {
					type="description",
					name=L["ASSISTANTS_DESC"],
					order=0,
				},
				
				learnassistants = {
					type="toggle",
					name=L["LEARNASSISTANTS_NAME"],
					desc=L["LEARNASSISTANTS_DESC"],
					order=10,
					get=function() return db.learnassistants end,
					set=function() db.learnassistants = not db.learnassistants end,
				},				

				newassistant = {
					type="input",
					name=L["NEWASSISTANT_NAME"],
					desc=L["NEWASSISTANT_DESC"],
					order=20,
					set=	function(info, val) table.insert(db.assistants, val) end,
				},
				removeassistant = {
					type="select",
					name=L["REMOVEASSISTANT_NAME"],
					desc=L["REMOVEASSISTANT_DESC"],
					order=21,
					values=	function() return db.assistants end,
					set=	function(i, v) table.remove(db.assistants, v) end,
				},
				
			}
		},

		groups = {
			type="group",
			name=L["GROUPSETUP_NAME"],
			args = {
				
				d = {
					type="description",
					name=L["GROUPSETUP_DESC"],
					order=0,
				},
				
				learnset = {
					type="toggle",
					name=L["LEARNSET_NAME"],
					desc=L["LEARNSET_DESC"],
					order=11,
					get=function() return db.learnset end,
					set=function() db.learnset = not db.learnset end,
				},				
				
				currentset = {
					type="select",
					name=L["CURRENTSET_NAME"],
					desc=L["CURRENTSET_DESC"],
					order=12,
					values=	function() return RaidGnome:GetSetNames() end,
					get=	function() return db.currentset end,
					set=	function(i, v) db.currentset = v end,
				},
				
				newset = {
					type="input",
					name=L["NEWSET_NAME"],
					desc=L["NEWSET_DESC"],
					order=20,
					set=	function(info, val) RaidGnome:CreateSet(val) end,
				},
				removeset = {
					type="select",
					name=L["REMOVESET_NAME"],
					desc=L["REMOVESET_DESC"],
					order=21,
					values=	function() return RaidGnome:GetSetNames() end,
					set=	function(i, v) 
								if db.currentset == v then
									db.currentset = nil
								end
								table.remove(db.sets, v)
							end,
				},
				
			}
			
		},
				
	}
	
}

function RaidGnome:GetPlayerScoreForGroup(cached, group)

	local bestmatch = 0
	
	-- For each member in the group
	for m, member in ipairs(group.members) do
		local match = 0
		
		-- Match on name?
		if cached.name == member.name then
			match = match + WEIGHT_NAME
		end
		
		-- Match on class?
		if cached.class == member.class then
			match = match + WEIGHT_CLASS
		end
		
		if match > bestmatch then
			bestmatch = match
		end
	end
	
	return bestmatch
end

function RaidGnome:AttemptToMovePlayer(cached)
	local set = self:GetSet()
	
	if set then

		-- For each group
		for g, group in ipairs(set.group) do
	
			local match = self:GetPlayerScoreForGroup(cached, group)
			
			self:Print(cached.name.." scored a match score of "..match.." for group "..g)
			
			-- If we have any sort of match...
			if match > 0 then
			
				-- Check how well the match is currently 
			
			end
				
		end
		
	end
end

-- Checks if player exists in group, matching on name
function RaidGnome:IsInGroup(name, group)
	for m, member in ipairs(group.members) do
		if member.name == name then
			return true
		end
	end
	return false
end

-- Removes a member from a given group, matching on name
function RaidGnome:RemoveFromGroup(name, group)
	for m, member in ipairs(group.members) do
		if member.name == name then
			table.remove(group.members, m)
		end
	end
end

-- Returns the current set
function RaidGnome:GetSet()
	return db.sets[db.currentset]
end

function RaidGnome:OpenOptions()
	InterfaceOptionsFrame_OpenToFrame(RaidGnome.optionsFrame)
end

function RaidGnome:SetLoot(method)
	if db.verbose then
		self:Print("Your RaidGnome tries to remember what loot type you wanted...")
	end
	if method == "master" then
		SetLootMethod(method, "player", 1)
	else
		SetLootMethod(method)	
	end
end

function RaidGnome:IsKnownAssistant(name)
	for index, assistant in pairs(db.assistants) do
		if assistant == name then
			return true
		end
	end
	return false
end

function RaidGnome:OnInitialize()
	TalentQuery.RegisterCallback(self, "TalentQuery_Ready")

	LibStub("AceConfig-3.0"):RegisterOptionsTable("RaidGnome", self.options)

	self:RegisterChatCommand("RaidGnome", "Command")
	self:RegisterChatCommand("rg", "Command")
	self:RegisterChatCommand("rgmove", "MovePlayerByName")
	self:RegisterChatCommand("rgtest", "ShowGroups")
	
	self.db = LibStub("AceDB-3.0"):New("RaidGnomeDB", defaults)
	db = self.db.char

	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("RaidGnome", "RaidGnome")
	
	self:CreateMMButton()
	self:SetMMButtonIcon("Interface\\Icons\\INV_Misc_Head_Gnome_01")
	self:SetMMButtonTooltip(L["BUTTONINFO1"],L["BUTTONINFO1"])
	self:ShowMMButton()
end

function RaidGnome:TalentQuery_Ready(e, name, realm)

end

function RaidGnome:ShowGroups(input)
	if db.currentset ~= nil then
		local set = self:GetSet()
		
		for g, group in ipairs(set.group) do
			self:Print("Group "..g)
			for m, member in ipairs(group.members) do
				self:Print("  name "..member.name..", class "..member.class)
			end
		end
		
	end
end

function RaidGnome:Command(input)
	self:OpenOptions()
end

function RaidGnome:CreateSet(name)
	-- Make sure there is no set with that name
	for i, set in ipairs(db.sets) do
		if db.sets[i].name == name then
			self:Print("There is already a set with that name.")
			return
		end
	end
	
	-- Add it to sets
	table.insert(db.sets, { ['name'] = name, ['group'] = {} })
end

function RaidGnome:OnMMButtonClick()
	self:OpenOptions()
end

function RaidGnome:OnEnable()
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	
	self:RegisterEvent("RAID_ROSTER_UPDATE")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
end

function RaidGnome:OnDisable()
end

function RaidGnome:MovePlayerByName(name)
	if db.currentset ~= nil then
			
		for i, cached in ipairs(raidcache) do
			if cached.name == name then
				self:AttemptToMovePlayer(cached)
				break
			end
		end

	end
	
end

function RaidGnome:UpdateRaid()
	local count = GetNumRaidMembers()
	for i=1, count do
		local name, rank, subgroup, level, class, fileName, zone, online, isDead, role, isML = GetRaidRosterInfo(i)
		
		-- Does member exist in raidcache?
		local cached = nil
		local new = true
		for index, member in pairs(raidcache) do
			if member.name == name then
				cached = member
				new = false
			end
		end
		
		-- Save to our internal cache, for determining changes
		if cached == nil then
			local newmember = {['name'] = name, ['rank'] = rank, ['subgroup'] = subgroup, ['class'] = class, ['isML'] = isML}
			cached = newmember
			table.insert(raidcache, newmember)
		end
		
		if new then
			-- Someone who just joined the raid

			if IsRaidLeader() then
			
				-- Promote known assistants
				if rank == 0 and self:IsKnownAssistant(name) then
					self:Print("Your RaidGnome remembers that you want "..name.." to be promoted, and does so.")
					PromoteToAssistant(name)
				end
				
			end
			
			if IsRaidLeader() or IsRaidOfficer() and db.currentset ~= nil then
			
				-- Move according to set
				--self:AttemptToMovePlayer(cached)
			
			end
		
		else
			-- Not a new joiner
			
			-- Auto-assigning raid assistants
			if IsRaidLeader() and db.learnassistants then
							
				-- Was member promoted to assistant?
				if rank == 1 and cached.rank == 0 and not self:IsKnownAssistant(name) then
					if db.verbose then
						self:Print("Your RaidGnome will remember to promote "..name.." next time.")
					end
					table.insert(db.assistants, name)
				end
				
				-- Was member demoted from assistant?
				if rank == 0 and cached.rank == 1 and self:IsKnownAssistant(name) then
					if db.verbose then
						self:Print("Your RaidGnome snickers and remembers not to promote "..name.." next time.")
					end
					for index, aname in pairs(db.assistants) do
						if name == aname then
							table.remove(db.assistants, index)
							break
						end
					end
				end
				
			end
		
			-- Remember group assignments
			if subgroup ~= cached.subgroup then
				
				if db.currentset ~= nil and db.learnset then
					
					local set = self:GetSet()
					
					if db.verbose then
						self:Print("Your RaidGnome will remember to make sure there is a similar "..class.." in group "..subgroup.." next time.")
					end
					
					-- Remove from old group if there is one (matching on name)
					if set.group[cached.subgroup] ~= nil then
						self:RemoveFromGroup(name, set.group[cached.subgroup])
					end
					
					-- Create the group if new
					if set.group[subgroup] == nil then
						set.group[subgroup] = {}
						set.group[subgroup].members = {}
					end
					
					-- Add our new member to the group with some basic attributes, if not already there
					if not self:IsInGroup(name, set.group[subgroup]) then
						table.insert(set.group[subgroup].members, {['name'] = name, ['class'] = class})
					end
					
					-- Trim the group if number of members > 5
					if table.getn(set.group[subgroup]) > 5 then
						table.remove(set.group[subgroup], 1)
					end
					
				end
				
				
			end
		
			-- Apply new values to our cached copy of member
			cached.subgroup = subgroup
			cached.rank = rank
			cached.isML = isML
		end
		
	end
		
end

function RaidGnome:PLAYER_ENTERING_WORLD()
	raidcache = {}
	
	if IsRaidLeader() or IsRaidOfficer() then
		self:UpdateRaid()
	end
end

function RaidGnome:PLAYER_TARGET_CHANGED()
	if IsRaidLeader() == 1 and GetNumRaidMembers() > 0 then
		if UnitClassification("target") == "boss" or UnitClassification("target") == "worldboss" and not UnitIsDead("target") then
			local lootmethod, masterlooterPartyID, masterlooterRaidID = GetLootMethod()
			if db.lootboss ~= lootmethod then
				-- Switch to boss method
				self:SetLoot(db.lootboss)
			end
		end
	end
end

function RaidGnome:RAID_ROSTER_UPDATE()
	-- Raid disbanded?
	if GetNumRaidMembers() == 0 then
		raidcache = {}
	end
end

function RaidGnome:PARTY_MEMBERS_CHANGED()
--	self:Print("PARTY_MEMBERS_CHANGED")
	if IsRaidLeader() or IsRaidOfficer() then
		self:UpdateRaid()
	end
end

function RaidGnome:PLAYER_REGEN_ENABLED()
	if IsRaidLeader() then
		local lootmethod, masterlooterPartyID, masterlooterRaidID = GetLootMethod()
	
		if db.lootdefault ~= lootmethod then
			-- Go back to last loot setting
			
			StaticPopupDialogs["DEFAULTLOOT_OPTION"] = {
				text = "Your RaidGnome thinks you may want to switch to the default loot option now. Is he correct?",
				button1 = OKAY,
				button2 = CANCEL,
				OnAccept = function() RaidGnome:SetLoot(db.lootdefault) end,
				timeout = 0,
				whileDead = 1,
				hideOnEscape = 1
			}
			StaticPopup_Show("DEFAULTLOOT_OPTION")
		end
	end
end

